import os.path
import requests
import json

import logging

logger = logging.getLogger("boschshcpy")

class SHCRegisterClient:
    """ Press and hold the button at the front of the SHC until the lights are blinking before you POST the command. When the SHC is not in pairing mode, there will be a connection error."""
    
    def __init__(self, controller_ip: str, password):
        """Initializes with IP address and access credentials."""
        self._controller_ip = controller_ip
        self._url = f"https://{self._controller_ip}:8443/smarthome/clients"

        # Settings for API call
        self._requests_session = requests.Session()
        self._requests_session.headers.update(
            {"Content-Type": "application/json",
             'Systempassword': password})
        self._requests_session.verify = False

        import urllib3
        urllib3.disable_warnings()

    def _post_api_or_fail(self, body):
        try:
            result = self._requests_session.post(self._url, data=json.dumps(body))
            if not result.ok:
                self._process_nok_result(result)
            if len(result.content) > 0:
                return json.loads(result.content)
            else:
                return {}
        except requests.exceptions.SSLError as e:
            logger.error('An error occured during requests.post(%s): %s. SHC not in pairing mode?' % (self._url, e))
            return {}

    def _process_nok_result(self, result):
        print(f"Body: {result.request.body}")
        print(f"Headers: {result.request.headers}")
        print(f"URL: {result.request.url}")
        raise ValueError(
            f"API call returned non-OK result (code {result.status_code})!: {result.content}")

    def register(self, id, name, certificate):
        if not os.path.exists(certificate):
            logger.error("No valid cert file, aborting!")
            # only continue if path to certificate is valid
            return None

        with open(certificate, 'r') as file:
            cert = file.read().replace('\n', '').replace('-----BEGIN CERTIFICATE-----', '-----BEGIN CERTIFICATE-----\r').replace('-----END CERTIFICATE-----', '\r-----END CERTIFICATE-----')

        data = {
            "@type": "client",
            "id": "oss_{}".format(id),
            "name": "OSS {}".format(name),
            "primaryRole": "ROLE_RESTRICTED_CLIENT",
            'certificate': cert
            }
        
        result = self._post_api_or_fail(data)
        return result['token'] if 'token' in result else None



from enum import Enum

from .device import SHCDevice
from .device_service import SHCDeviceService
from .services_impl import ShutterContactService, ShutterControlService, CameraLightService, IntrusionDetectionControlService


class SHCSmokeDetector(SHCDevice):
    from .services_impl import AlarmService, SmokeDetectorCheckService

    class AlarmState(Enum):
        INTRUSION_ALARM_ON_REQUESTED = "INTRUSION_ALARM_ON_REQUESTED"
        INTRUSION_ALARM_OFF_REQUESTED = "INTRUSION_ALARM_OFF_REQUESTED"
        MUTE_SECONDARY_ALARM_REQUESTED = "MUTE_SECONDARY_ALARM_REQUESTED"

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device)

        self._alarm_service: AlarmService = self.device_service('Alarm')
        self._smokedetectorcheck_service: SmokeDetectorCheckService = self.device_service(
            'SmokeDetectorCheck')

    def set_alarmstate(self, state: AlarmState):
        self._alarm_service.put_state_element('state', state.name)

    @property
    def alarm_state(self) -> AlarmService.State:
        return self._alarm_service.value

    @property
    def smokedetectorcheck_state(self) -> SmokeDetectorCheckService.State:
        return self._smokedetectorcheck_service.value

    def update(self):
        self._alarm_service.short_poll()
        self._smokedetectorcheck_service.short_poll()

    def summary(self):
        print(f"SD SmokeDetector:")
        super().summary()


class SHCSmartPlug(SHCDevice):
    from .services_impl import PowerSwitchService, PowerMeterService, PowerSwitchProgramService

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device)

        self._powerswitch_service = self.device_service('PowerSwitch')
        self._powerswitchprogram_service = self.device_service('PowerSwitchProgram')
        self._powermeter_service = self.device_service('PowerMeter')

    def set_state(self, state: bool):
        self._powerswitch_service.put_state_element('switchState', "ON" if state else "OFF")

    @property
    def state(self) -> PowerSwitchService.State:
        return self._powerswitch_service.value

    @property
    def energyconsumption(self):
        return self._powermeter_service.energyconsumption

    @property
    def powerconsumption(self):
        return self._powermeter_service.powerconsumption

    @property
    def state(self) -> PowerSwitchService.State:
        return self._powerswitch_service.value

    def update(self):
        self._powerswitch_service.short_poll()
        self._powerswitchprogram_service.short_poll()
        self._powermeter_service.short_poll()

    def summary(self):
        print(f"PSM SmartPlug:")
        super().summary()


class SHCShutterControl(SHCDevice):
    from .services_impl import ShutterControlService

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device=raw_device)
        self._service: ShutterControlService = self.device_service('ShutterControl')

    def set_level(self, level):
        self._service.put_state_element('level', level)

    @property
    def level(self) -> float:
        return self._service.level

    def set_stopped(self):
        self._service.put_state_element('operationState', ShutterControlService.State.STOPPED.name)

    @property
    def operation_state(self) -> ShutterControlService.State:
        return self._service.value

    def update(self):
        self._service.short_poll()

    def summary(self):
        print(f"BBL ShutterControl:")
        super().summary()


class SHCShutterContact(SHCDevice):
    from .services_impl import ShutterContactService
    class DeviceClass(Enum):
        GENERIC = "GENERIC"
        ENTRANCE_DOOR = "ENTRANCE_DOOR"
        REGULAR_WINDOW = "REGULAR_WINDOW"
        FRENCH_WINDOW = "FRENCH_WINDOW"

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device)
        self._service = self.device_service('ShutterContact')

    @property
    def device_class(self) -> DeviceClass:
        return self.DeviceClass(self.profile)

    @property
    def state(self) -> ShutterContactService.State:
        return self._service.value

    def update(self):
        self._service.short_poll()

    def summary(self):
        print(f"SWD ShutterContact:")
        super().summary()


class SHCCameraEyes(SHCDevice):
    from .services_impl import CameraLightService, CameraNotificationService, PrivacyModeService

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device)

        self._privacymode_service = self.device_service('PrivacyMode')
        self._cameranotification_service = self.device_service(
            'CameraNotification')
        self._cameralight_service = self.device_service('CameraLight')

    def set_privacymode(self, state: bool):
        self._privacymode_service.put_state_element(
            'value', "ENABLED" if state else "DISABLED")

    def set_cameranotification(self, state: bool):
        self._cameranotification_service.put_state_element(
            'value', "ENABLED" if state else "DISABLED")

    def set_cameralight(self, state: bool):
        self._cameralight_service.put_state_element(
            'value', "ON" if state else "OFF")

    @property
    def lightstate(self) -> CameraLightService.State:
        return self._cameralight_service.value

    def update(self):
        self._cameralight_service.short_poll()
        self._cameranotification_service.short_poll()
        self._privacymode_service.short_poll()

    def summary(self):
        print(f"CAMERA_EYES CameraEyes:")
        super().summary()


class SHCIntrusionDetectionSystem(SHCDevice):
    from .services_impl import IntrusionDetectionControlService

    def __init__(self, api, raw_device):
        super().__init__(api, raw_device)
        self._service = self.device_service('IntrusionDetectionControl')

    def set_alarmstate(self, state: IntrusionDetectionControlService.State):
        self._service.put_state_element('value', state.name)

    def disarm(self):
        self.set_alarmstate(IntrusionDetectionControlService.State.SYSTEM_DISARMED)

    def arm(self):
        self.set_alarmstate(IntrusionDetectionControlService.State.SYSTEM_ARMED)

    def arm_activation_delay(self, seconds):
        if self.alarmstate == IntrusionDetectionControlService.State.SYSTEM_ARMING:
            return

        self._service.put_state_element('armActivationDelayTime', seconds)

    def arm_instant(self):
        if self.alarmstate == IntrusionDetectionControlService.State.SYSTEM_ARMING:
            return

        delay_time = self._service.armActivationDelayTime
        self.arm_activation_delay(1)
        self.arm()
        self.arm_activation_delay(delay_time)

    def mute_alarm(self):
        self.set_alarmstate(IntrusionDetectionControlService.State.MUTE_ALARM)

    def trigger(self):
        # not implemented yet
        print("Trigger alarm simulation")

    @property
    def alarmstate(self) -> IntrusionDetectionControlService.State:
        return self._service.value

    @property
    def armActivationDelayTime(self):
        return self._service.value

    def update(self):
        self._service.short_poll()

    def summary(self):
        print(f"-IntrusionDetectionSystem-:")
        super().summary()


MODEL_MAPPING = {
    "SWD": "ShutterContact",
    "BBL": "ShutterControl",
    "PSM": "SmartPlug",
    "BSM": "LightControl", # uses same impl as PSM
    "SD": "SmokeDetector",
    "CAMERA_EYES": "CameraEyes",
    "INTRUSION_DETECTION_SYSTEM": "-IntrusionDetectionSystem-",
}
# "WRC2": "UniversalSwitchFlex",
# "MD": "MotionDetector"

SUPPORTED_MODELS = MODEL_MAPPING.keys()



import typing
import logging

from .device_service import SHCDeviceService
from .services_impl import build, SUPPORTED_DEVICE_SERVICE_IDS

logger = logging.getLogger("boschshcpy")


class SHCDevice:
    def __init__(self, api, raw_device):
        self._api = api
        self._raw_device = raw_device

        self._device_services_by_id = {}
        self._enumerate_services()

    def _enumerate_services(self):
        for device_service_id in self._raw_device['deviceServiceIds']:
            if device_service_id not in SUPPORTED_DEVICE_SERVICE_IDS:
                continue

            raw_device_service_data = self._api.get_device_service(self.id, device_service_id)
            device_service = build(self._api, raw_device_service_data)

            self._device_services_by_id[device_service_id] = device_service

    @property
    def root_device_id(self):
        return self._raw_device['rootDeviceId']

    @property
    def id(self):
        return self._raw_device['id']

    @property
    def manufacturer(self):
        return self._raw_device['manufacturer']

    @property
    def room_id(self):
        return self._raw_device['roomId'] if 'roomId' in self._raw_device else None

    @property
    def device_model(self):
        return self._raw_device['deviceModel']

    @property
    def serial(self):
        return self._raw_device['serial'] if 'serial' in self._raw_device else None

    @property
    def profile(self):
        return self._raw_device['profile'] if 'profile' in self._raw_device else None

    @property
    def name(self):
        return self._raw_device['name']

    @property
    def status(self):
        return self._raw_device['status']

    @property
    def device_services(self) -> typing.Sequence[SHCDeviceService]:
        return list(self._device_services_by_id.values())

    @property
    def device_service_ids(self) -> typing.Set[str]:
        return set(self._device_services_by_id.keys())
    
    def device_service(self, device_service_id):
        return self._device_services_by_id[device_service_id]

    def summary(self):
        print(f"Device: {self.id}")
        print(f"  Name          : {self.name}")
        print(f"  Manufacturer  : {self.manufacturer}")
        print(f"  Model         : {self.device_model}")
        print(f"  Room          : {self.room_id}")
        print(f"  Serial        : {self.serial}")
        print(f"  Profile       : {self.profile}")
        for device_service in self.device_services:
            device_service.summary()

    def process_long_polling_poll_result(self, raw_result):
        assert raw_result["@type"] == "DeviceServiceData"
        device_service_id = raw_result["id"]
        if device_service_id in self._device_services_by_id.keys():
            device_service = self._device_services_by_id[device_service_id]
            device_service.process_long_polling_poll_result(raw_result)
        else:
            logger.debug(f"Skipping polling result with unknown device service id {device_service_id}.")



from .session import SHCSession
from .scenario import SHCScenario
from .device_helper import SHCDeviceHelper, SHCCameraEyes, SHCShutterContact, SHCShutterControl, SHCSmartPlug, SHCSmokeDetector, SHCIntrusionDetectionSystem



import typing
import logging
import threading
import time
import sys

from .api import SHCAPI, JSONRPCError
from .device import SHCDevice
from .room import SHCRoom
from .scenario import SHCScenario
from .information import SHCInformation
from .services_impl import SUPPORTED_DEVICE_SERVICE_IDS
from .device_helper import SHCDeviceHelper

logger = logging.getLogger("boschshcpy")


class SHCSession:
    def __init__(self, controller_ip: str, certificate, key, enumerate = True):
        # API
        self._api = SHCAPI(controller_ip=controller_ip, certificate=certificate, key=key)
        self._device_helper = SHCDeviceHelper(self._api)

        # Subscription status
        self._poll_id = None

        # SHC Information
        self._shc_information = None

        # All devices
        self._rooms_by_id = {}
        self._scenarios_by_id = {}
        self._devices_by_id = {}

        self._get_information()

        if enumerate: self._enumerate_all()

        self._polling_thread = None
        self._stop_polling_thread = False

    def _enumerate_all(self):
        self._enumerate_devices()
        self._enumerate_rooms()
        self._enumerate_scenarios()
        self._get_information()

    def _enumerate_devices(self):
        raw_devices = self._api.get_devices()

        for raw_device in raw_devices:
            device_id = raw_device['id']

            if set(raw_device['deviceServiceIds']).isdisjoint(SUPPORTED_DEVICE_SERVICE_IDS):
                logger.debug(f"Skipping device id {device_id} which has no services that are supported by this library")
                continue
            
            self._devices_by_id[device_id] = self._device_helper.device_init(raw_device)

    def _enumerate_rooms(self):
        raw_rooms = self._api.get_rooms()
        for raw_room in raw_rooms:
            room_id = raw_room["id"]
            room = SHCRoom(api=self._api, raw_room=raw_room)
            self._rooms_by_id[room_id] = room

    def _enumerate_scenarios(self):
        raw_scenarios = self._api.get_scenarios()
        for raw_scenario in raw_scenarios:
            scenario_id = raw_scenario["id"]
            scenario = SHCScenario(api=self._api, raw_scenario=raw_scenario)
            self._scenarios_by_id[scenario_id] = scenario

    def _get_information(self):
        raw_information = self._api.get_shcinformation()
        self._shc_information = SHCInformation(api=self._api, raw_information=raw_information)

    def _long_poll(self, wait_seconds=30):
        if self._poll_id is None:
            self._poll_id = self.api.long_polling_subscribe()
            logger.debug(f"Subscribed for long poll. Poll id: {self._poll_id}")
        try:
            raw_results = self.api.long_polling_poll(self._poll_id, wait_seconds)
            for raw_result in raw_results:
                self._process_long_polling_poll_result(raw_result)

            return True
        except JSONRPCError as json_rpc_error:
            if json_rpc_error.code == -32001:
                self._poll_id = None
                logger.warning(f"SHC claims unknown poll id. Invalidating poll id and trying resubscribe next time...")
                return False
            else:
                raise json_rpc_error

    def _maybe_unsubscribe(self):
        if self._poll_id is not None:
            self.api.long_polling_unsubscribe(self._poll_id)

    def _process_long_polling_poll_result(self, raw_result):
        if raw_result["@type"] != "DeviceServiceData":
            # Skipping polling results of type message, device or unknown type
            return
        device_id = raw_result["deviceId"]
        if device_id in self._devices_by_id.keys():
            device = self._devices_by_id[device_id]
            device.process_long_polling_poll_result(raw_result)
        else:
            logger.debug(f"Skipping polling result with unknown device id {device_id}.")
    
    def start_polling(self):
        if self._polling_thread is None:
            def polling_thread_main():
                while not self._stop_polling_thread:
                    try:
                        if not self._long_poll():
                            logging.warning("_long_poll returned False. Waiting 1 second.")
                            time.sleep(1.0)
                    except RuntimeError as err:
                        self._stop_polling_thread = True
                        logging.info("Stopping polling thread after expected runtime error.")
                        logging.info(f"Error description: {err}. {err.args}")
                        logging.info(f"Attempting unsubscribe...")
                        try:
                            self._maybe_unsubscribe()
                        except Exception as ex:
                            logging.info(f"Unsubscribe unsuccessful: {ex}")

                    except Exception as ex:
                        logging.error(f"Error in polling thread: {ex}. Waiting 15 seconds.")
                        time.sleep(15.0)

            self._polling_thread = threading.Thread(target=polling_thread_main, name="SHCPollingThread")
            self._polling_thread.start()

        else:
            raise ValueError("Already polling!")

    def stop_polling(self):
        if self._polling_thread is not None:
            self._stop_polling_thread = True
            self._polling_thread.join()

            self._maybe_unsubscribe()
            self._polling_thread = None
            self._poll_id = None
        else:
            raise ValueError("Not polling!")

    def reinitialize(self):
        reinit_polling = False
        if self._polling_thread is not None:
            self.stop_polling()
            reinit_polling = True

        self._shc_information = None

        self._rooms_by_id = {}
        self._scenarios_by_id = {}
        self._devices_by_id = {}

        self._get_information()
        self._enumerate_all()

        if reinit_polling:
            self.start_polling()

    @property
    def devices(self) -> typing.Sequence[SHCDevice]:
        return list(self._devices_by_id.values())

    def device(self, device_id) -> SHCDevice:
        return self._devices_by_id[device_id]

    @property
    def rooms(self) -> typing.Sequence[SHCRoom]:
        return list(self._rooms_by_id.values())

    def room(self, room_id) -> SHCRoom:
        return self._rooms_by_id[room_id]

    @property
    def scenarios(self) -> typing.Sequence[SHCScenario]:
        return list(self._scenarios_by_id.values())

    def scenario(self, scenario_id) -> SHCScenario:
        return self._scenarios_by_id[scenario_id]

    @property
    def information(self) -> SHCInformation:
        return self._shc_information
    
    @property
    def api(self):
        return self._api

    @property
    def device_helper(self):
        return self._device_helper



import typing
import logging

from .device import SHCDevice
from .models_impl import SUPPORTED_MODELS, SHCCameraEyes, SHCShutterContact, SHCShutterControl, SHCSmartPlug, SHCSmokeDetector, SHCIntrusionDetectionSystem

logger = logging.getLogger("boschshcpy")


class SHCDeviceHelper:
    def __init__(self, api):
        self._api = api
        self._devices_by_model = {}
        for model in SUPPORTED_MODELS:
            self._devices_by_model[model] = {}

    def device_init(self, raw_device):
        device_id = raw_device['id']
        device_model = raw_device['deviceModel']

        device = None
        switcher = {
            'SWD': lambda: SHCShutterContact(api=self._api, raw_device=raw_device),
            'BBL': lambda: SHCShutterControl(api=self._api, raw_device=raw_device),
            'PSM': lambda: SHCSmartPlug(api=self._api, raw_device=raw_device),
            'BSM': lambda: SHCSmartPlug(api=self._api, raw_device=raw_device),
            'SD': lambda: SHCSmokeDetector(api=self._api, raw_device=raw_device),
            'CAMERA_EYES': lambda: SHCCameraEyes(api=self._api, raw_device=raw_device),
            'INTRUSION_DETECTION_SYSTEM': lambda: SHCIntrusionDetectionSystem(api=self._api, raw_device=raw_device)
        }
        if device_model in switcher and device_model in SUPPORTED_MODELS:
            device = switcher[device_model]()
            self._devices_by_model[device_model][device_id] = device
        else:
            device = SHCDevice(api=self._api, raw_device=raw_device)

        return device

    @property
    def shutter_contacts(self) -> typing.Sequence[SHCShutterContact]:
        if 'SWD' not in SUPPORTED_MODELS: 
            return []
        return list(self._devices_by_model['SWD'].values())

    @property
    def shutter_controls(self) -> typing.Sequence[SHCShutterControl]:
        if 'BBL' not in SUPPORTED_MODELS:
            return []
        return list(self._devices_by_model['BBL'].values())

    @property
    def smart_plugs(self) -> typing.Sequence[SHCSmartPlug]:
        if 'PSM' not in SUPPORTED_MODELS:
            return []
        return list(self._devices_by_model['PSM'].values())

    @property
    def light_controls(self) -> typing.Sequence[SHCSmartPlug]:
        if 'BSM' not in SUPPORTED_MODELS:
            return []
        return list(self._devices_by_model['BSM'].values())

    @property
    def smoke_detectors(self) -> typing.Sequence[SHCSmokeDetector]:
        if 'SD' not in SUPPORTED_MODELS:
            return []
        return list(self._devices_by_model['SD'].values())

    @property
    def camera_eyes(self) -> typing.Sequence[SHCCameraEyes]:
        if 'CAMERA_EYES' not in SUPPORTED_MODELS:
            return []
        return list(self._devices_by_model['CAMERA_EYES'].values())

    @property
    def intrusion_detection_system(self) -> SHCIntrusionDetectionSystem:
        if 'INTRUSION_DETECTION_SYSTEM' not in SUPPORTED_MODELS:
            return None
        return list(self._devices_by_model['INTRUSION_DETECTION_SYSTEM'].values())[0]



class SHCRoom:
    def __init__(self, api, raw_room):
        self._api = api
        self._raw_room = raw_room

    @property
    def id(self):
        return self._raw_room["id"]

    @property
    def icon_id(self):
        return self._raw_room["iconId"]

    @property
    def name(self):
        return self._raw_room["name"]

    def summary(self):
        print(f"Room: {self.id}")
        print(f"  Name   : {self.name}")
        print(f"  Icon Id: {self.icon_id}")



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="boschshcpy",
    version="0.1.2",
    url="https://github.com/tschamm/boschshcpy",
    author="Clemens-Alexander Brust, Thomas Schamm",
    author_email="cabrust@pm.me, thomas@tschamm.de",
    description="Bosch Smart Home Controller API Python Library",
    license='bsd-3-clause',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "License :: OSI Approved :: BSD License"
    ],
    platform='any',
    python_requires='>=3.7',
    install_requires=['requests>=2.22']
)



from enum import Enum


class SHCInformation:
    class UpdateState(Enum):
        NO_UPDATE_AVAILABLE = "NO_UPDATE_AVAILABLE"
        DOWNLOADING = "DOWNLOADING"
        UPDATE_IN_PROGRESS = "UPDATE_IN_PROGRESS"
        UPDATE_AVAILABLE = "UPDATE_AVAILABLE"

    def __init__(self, api, raw_information):
        self._api = api
        self._raw_information = raw_information

    @property
    def version(self):
        return self._raw_information["version"] if self._raw_information else "n/a"

    @property
    def updateState(self) -> UpdateState:
        return self.UpdateState(self._raw_information["updateState"] if self._raw_information else "NO_UPDATE_AVAILABLE")

    @property
    def connectivityVersion(self):
        return self._raw_information["connectivityVersion"] if self._raw_information else "n/a"

    def summary(self):
        print(f"Information:")
        print(f"  SW Version         : {self.version}")
        print(f"  State              : {self.updateState}")
        print(f"  connectivityVersion: {self.connectivityVersion}")
